/****************************************************************************\
 *
 * magnetiX - Magnetic Scrolls Interpreter - Mac OS X port
 *
 * Written by Jan-Sebastian Schliemann <magnetiX@maczentrisch.de>
 *
 *
 * the original Magnetic was:
 *
 * Written by Niclas Karlsson <nkarlsso@abo.fi>,
 *            David Kinder <davidk.kinder@virgin.net>,
 *            Stefan Meier <Stefan.Meier@if-legends.org> and
 *            Paul David Doherty <pdd@if-legends.org>
 *
 * Copyright (C) 1997-2008  Niclas Karlsson
 *
 *
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 *
 \*****************************************************************************/

#import "MXScrollView.h"
#import "MXTextView.h"
#import "MagneticController.h"
#import "GameSelectController.h"

@implementation MXScrollView


- (void)awakeFromNib
{
    oldTargetPoint = NSMakePoint(0, 0);
    
    [[self contentView] setPostsBoundsChangedNotifications:YES];
	
	self.scrollerTarget = self.verticalScroller.target;
	self.scrollerAction = self.verticalScroller.action;
	
	self.verticalScroller.target = self;
	self.verticalScroller.action = @selector(optimizedScrollerAction:);
	
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(synchronizedViewContentBoundsDidChange:) name:NSViewBoundsDidChangeNotification object:nil];
}


- (void)optimizedScrollerAction:(NSScroller *)sender
{
	if (NSApp.currentEvent.type != NSEventTypeLeftMouseDown) {
		[sender sendAction:self.scrollerAction to:self.scrollerTarget];
		return;
	}
	
	switch (sender.hitPart) {
		case NSScrollerDecrementPage:
			[self pageUp:self];
			break;
			
		case NSScrollerIncrementPage:
			[self pageDown:self];
			break;
			
		default:
			[sender sendAction:self.scrollerAction to:self.scrollerTarget];
			break;
	}
}


- (void)drawRect:(NSRect)dirtyRect // prevents redraw issue with visible scrollers!
{
	[NSColor.clearColor set];
	NSRectFill(self.bounds);
}


- (void)setFrame:(NSRect)frameRect
{
	MagneticController *controller = [(GameSelectController *)[[self window] delegate] magneticController];
	CGFloat maxWidth = [controller currentContentWidth];

	NSRect frame = [[self window] frame];
	
	frameRect.size.width = ceilf(((frame.size.width - maxWidth) / 2) + maxWidth);
	frameRect.origin.x = frame.size.width - frameRect.size.width;
	
	[super setFrame:frameRect];
}


- (void)updateState
{
	if (![_timer isValid]) {
		self.timer = [NSTimer timerWithTimeInterval:0.1 target:self selector:@selector(updateStateForReal) userInfo:nil repeats:NO];
		[[NSRunLoop currentRunLoop] addTimer:self.timer forMode:NSRunLoopCommonModes];
	}
}


- (void)updateStateForReal
{
	MagneticController *controller = [(GameSelectController *)[[self window] delegate] magneticController];
	
	if ([controller isScriptPlaying]) { return; }
	
	id attributes = [[self documentView] attributesForLocation];
	
	NSString *cacheKey = attributes[@"MXImageShown"];
	id status = attributes[@"MXStatusShown"];
	
	[controller mainRestoreStatus:status];
	[controller showImage:cacheKey];
}


- (void)synchronizedViewContentBoundsDidChange:(NSNotification *)notification
{
	[self updateState];
	
	id contentView = [self contentView];
	NSPoint targetPoint = [contentView bounds].origin;
	
	if (targetPoint.y >= [[self documentView] bounds].size.height - [contentView bounds].size.height - 1) {
		[[(GameSelectController *)[[self window] delegate] magneticController] disableMore];
	}
    
    if (!NSEqualPoints(oldTargetPoint, targetPoint)) {
        oldTargetPoint = targetPoint;
        [self reflectScrolledClipView:contentView];
    } else {
        oldTargetPoint = targetPoint;
    }
}


- (void)scrollWheel:(NSEvent *)theEvent
{
	if (!self.isScrollingPrevented) {
		[super scrollWheel:theEvent];
	}
}


- (void)pageUp:(id)sender
{
	MagneticController *controller = [(GameSelectController *)[[self window] delegate] magneticController];
	
	NSRect contentBounds = [[self contentView] bounds];
	
	NSPoint targetPoint = contentBounds.origin;
	
	CGFloat theHeight = controller.topGradient.frame.origin.y - controller.botGradient.frame.origin.y;
	
	targetPoint.y -= (theHeight - [controller gradientHeight] * 0.5);
	
	if (targetPoint.y < 0) {
		targetPoint.y = 0;
	}
	
	[self restrictedScrollToPoint:targetPoint];
}


- (void)pageDown:(id)sender
{
	MagneticController *controller = [(GameSelectController *)[[self window] delegate] magneticController];

	NSRect contentBounds = [[self contentView] bounds];
	NSRect documentBounds = [[self documentView] bounds];
	
	NSPoint targetPoint = contentBounds.origin;
	
	CGFloat theHeight = controller.topGradient.frame.origin.y - controller.botGradient.frame.origin.y;

	targetPoint.y += (theHeight - [controller gradientHeight] * 0.5);
	
	if (targetPoint.y > documentBounds.size.height - contentBounds.size.height) {
		targetPoint.y = documentBounds.size.height - contentBounds.size.height;
	}
	
	[self restrictedScrollToPoint:targetPoint];
}


- (void)moveToBeginningOfDocument:(nullable id)sender
{
	[self restrictedScrollToPoint:NSMakePoint(0.0, 0.0)];
}


- (void)moveToEndOfDocument:(nullable id)sender
{
	NSPoint p = NSZeroPoint;
	p.y = self.documentView.bounds.size.height - self.contentView.bounds.size.height;
	[self restrictedScrollToPoint:p];
}


- (void)restrictedScrollToPoint:(NSPoint)point
{	
	[NSAnimationContext beginGrouping];
	[[[self contentView] animator] setBoundsOrigin:point];
	[self flashScrollers];
	[NSAnimationContext endGrouping];
}


+ (BOOL)isCompatibleWithResponsiveScrolling
{
	return YES;
}



- (void)dealloc
{
	self.timer = nil;
	self.scrollerTarget = nil;
	
	[super dealloc];
}


@end
